import json
import logging
import time
from dataclasses import asdict
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Type, Union
import tornado
import tornado.ioloop
import tornado.web
from pydantic import BaseModel
from jupyter_core.utils import ensure_async
from jupyter_server.base.handlers import JupyterHandler
from tornado.websocket import WebSocketHandler
from mito_ai.message_history import GlobalMessageHistory
from mito_ai.logger import get_logger
from mito_ai.models import (
    AgentMessageBuilder,
    ChatMessageBuilder,
    IncomingMessageTypes,
    CodeExplainMessageBuilder,
    CompletionError,
    CompletionItem,
    CompletionReply,
    CompletionRequest,
    CompletionStreamChunk,
    ErrorMessage,
    FetchHistoryReply,
    InlineCompletionMessageBuilder,
    SmartDebugMessageBuilder
)
from mito_ai.prompt_builders.smart_debug_prompt import remove_inner_thoughts_from_message
from mito_ai.providers import OpenAIProvider
from mito_ai.utils.create import initialize_user
from mito_ai.providers import OpenAIProvider
from mito_ai.utils.create import initialize_user
from mito_ai.utils.version_utils import is_pro
from openai.types.chat import ChatCompletionMessageParam

__all__ = ["CompletionHandler"]

# Global history instance
message_history = GlobalMessageHistory()

# This handler is responsible for the mito-ai/completions endpoint.
# It takes a message from the user, sends it to the OpenAI API, and returns the response.
# Important: Because this is a server extension, print statements are sent to the
# jupyter server terminal by default (ie: the terminal you ran `jupyter lab`)
class CompletionHandler(JupyterHandler, WebSocketHandler):
    """Completion websocket handler."""

    def initialize(self, llm: OpenAIProvider) -> None:
        super().initialize()
        self.log.debug("Initializing websocket connection %s", self.request.path)
        self._llm = llm
        self.full_message_history: list[ChatCompletionMessageParam] = []
        self.is_pro = is_pro()

    @property
    def log(self) -> logging.Logger:
        """Use Mito AI logger"""
        return get_logger()

    @tornado.web.authenticated
    def head(self) -> None:
        """Handle a HEAD request for the websocket."""
        self.set_status(HTTPStatus.OK)
        self.finish()

    async def pre_get(self) -> None:
        """Handles websocket authentication/authorization."""
        # authenticate the request before opening the websocket
        user = self.current_user
        if user is None:
            self.log.warning("Couldn't authenticate WebSocket connection")
            raise tornado.web.HTTPError(HTTPStatus.UNAUTHORIZED)

        # authorize the user.
        if not await ensure_async(
            self.authorizer.is_authorized(self, user, "execute", "mito-ai-completion")
        ):
            raise tornado.web.HTTPError(HTTPStatus.FORBIDDEN)

    async def get(self, *args: Any, **kwargs: dict[str, Any]) -> None:
        """Get an event to open a socket."""
        # This method ensure to call `pre_get` before opening the socket.
        await ensure_async(self.pre_get()) # type: ignore

        initialize_user()

        reply = super().get(*args, **kwargs)
        if reply is not None:
            await reply

    def on_close(self) -> None:
        """Invoked when the WebSocket is closed.

        If the connection was closed cleanly and a status code or reason
        phrase was supplied, these values will be available as the attributes
        ``self.close_code`` and ``self.close_reason``.
        """
        # Stop observing the provider error
        self._llm.unobserve(self._send_error, "last_error")
        
        # Clear the message history
        self.full_message_history = []
        

    async def on_message(self, message: str) -> None: # type: ignore
        """Handle incoming messages on the WebSocket.

        Args:
            message: The message received on the WebSocket.
        """
        
        # first, verify that the message is an `CompletionRequest`.
        self.log.debug("Message received: %s", message)
        try:
            parsed_message = json.loads(message)

            metadata_dict = parsed_message.get('metadata', {})
            type: IncomingMessageTypes = parsed_message.get('type')
        except ValueError as e:
            self.log.error("Invalid completion request.", exc_info=e)
            return

        # Clear history if the type is "clear_history"
        if type == "clear_history":
            message_history.clear_histories()
            return
        
        if type == "fetch_history":
            _, display_history = message_history.get_histories()
            reply = FetchHistoryReply(
                parent_id=parsed_message.get('message_id'),
                items=display_history
            )
            
            self.reply(reply)
            return
        
        response_format = None

        # Generate new message based on message type
        if type == "inline_completion":
            # Reset the message history for the inline completion 
            # because they are ephemeral
            self.full_message_history = []
            
            inlineCompletionPromptBuilder = InlineCompletionMessageBuilder(**metadata_dict)
            prompt = inlineCompletionPromptBuilder.prompt
            model = inlineCompletionPromptBuilder.pro_model if self.is_pro else inlineCompletionPromptBuilder.os_model
            
            ai_optimized_history: List[ChatCompletionMessageParam] = [{"role": "user", "content": prompt}]
            
        else:
            if type == "chat" or type == "agent:execution":
                chatMessagePromptBuilder = ChatMessageBuilder(**metadata_dict)
                prompt = chatMessagePromptBuilder.prompt
                display_message = chatMessagePromptBuilder.display_message
                model = chatMessagePromptBuilder.pro_model if self.is_pro else chatMessagePromptBuilder.os_model

                if chatMessagePromptBuilder.index is not None:
                    # Clear the chat history after the specified index (inclusive)
                    self.full_message_history = self.full_message_history[:chatMessagePromptBuilder.index]

            elif type == "codeExplain":
                codeExplainPromptBuilder = CodeExplainMessageBuilder(**metadata_dict)
                prompt = codeExplainPromptBuilder.prompt
                display_message = codeExplainPromptBuilder.display_message
                model = codeExplainPromptBuilder.pro_model if self.is_pro else codeExplainPromptBuilder.os_model
            elif type == "smartDebug":
                smartDebugPromptBuilder = SmartDebugMessageBuilder(**metadata_dict)
                prompt = smartDebugPromptBuilder.prompt
                display_message = smartDebugPromptBuilder.display_message
                model = smartDebugPromptBuilder.pro_model if self.is_pro else smartDebugPromptBuilder.os_model
            elif type == "agent:planning":
                agentMessageBuilder = AgentMessageBuilder(**metadata_dict)
                prompt = agentMessageBuilder.prompt
                display_message = agentMessageBuilder.display_message
                model = agentMessageBuilder.pro_model if self.is_pro else agentMessageBuilder.os_model
                response_format = agentMessageBuilder.response_format
            else:
                raise ValueError(f"Invalid message type: {type}")

            new_ai_optimized_message: ChatCompletionMessageParam = {"role": "user", "content": prompt}
            new_display_message: ChatCompletionMessageParam = {"role": "user", "content": display_message}
            message_history.append_message(new_ai_optimized_message, new_display_message)
            ai_optimized_history, display_history = message_history.get_histories()

        request = CompletionRequest(
            type=type,
            message_id=parsed_message.get('message_id'),
            messages=ai_optimized_history,
            stream=parsed_message.get('stream', False)
        )

        try:
            if request.stream and self._llm.can_stream:
                await self._handle_stream_request(request, prompt_type=request.type, model=model)
            else:
                await self._handle_request(
                    request,
                    prompt_type=request.type,
                    model=model,
                    response_format=response_format
                )
        except Exception as e:
            await self.handle_exception(e, request)

    def open(self, *args: str, **kwargs: str) -> None:
        """Invoked when a new WebSocket is opened.

        The arguments to `open` are extracted from the `tornado.web.URLSpec`
        regular expression, just like the arguments to
        `tornado.web.RequestHandler.get`.

        `open` may be a coroutine. `on_message` will not be called until
        `open` has returned.
        """
        if self._llm.last_error:
            self._send_error({"new": self._llm.last_error})
        # Start observing the provider error
        self._llm.observe(self._send_error, "last_error")
        # Send the server capabilities to the client.
        self.reply(self._llm.capabilities)

    async def handle_exception(self, e: Exception, request: CompletionRequest) -> None:
        """
        Handles an exception raised in either ``handle_request`` or
        ``handle_stream_request``.

        Args:
            e: The exception raised.
            request: The completion request that caused the exception.
        """
        hint = ""
        if isinstance(e, PermissionError):
            hint = "You've reached the free tier limit for Mito AI. Upgrade to Pro for unlimited uses or supply your own OpenAI API key."
        elif "openai" in self._llm.capabilities.provider.lower():
            hint = "There was an error communicating with OpenAI. This might be due to a temporary OpenAI outage, a problem with your internet connection, or an incorrect API key. Please try again."
        else:
            hint = "There was an error communicating with Mito server. This might be due to a temporary server outage or a problem with your internet connection. Please try again."
        
        error: CompletionError = CompletionError.from_exception(e, hint=hint)
        self._send_error({"new": error})
        
        reply: Union[CompletionStreamChunk, CompletionReply]
        if request.stream:
            reply = CompletionStreamChunk(
                chunk=CompletionItem(content="", isIncomplete=True),
                parent_id=request.message_id,
                done=True,
                error=error,
            )
        else:
            reply = CompletionReply(
                items=[],
                error=error,
                parent_id=request.message_id,
            )
        self.reply(reply)

    async def _handle_request(
        self,
        request: CompletionRequest,
        prompt_type: str,
        model: str,
        response_format: Optional[Type[BaseModel]] = None,
    ) -> None:
        """Handle completion request.

        Args:
            request: The completion request description.
        """
        start = time.time()
        reply = await self._llm.request_completions(request, prompt_type, model, response_format)

        # Save to the message history
        # Inline completion is ephemeral and does not need to be saved
        if request.type != "inline_completion":
            response = reply.items[0].content if reply.items else ""

            ai_optimized_message: ChatCompletionMessageParam = {
                "role": "assistant", 
                "content": response
            }
            display_message: ChatCompletionMessageParam = {
                "role": "assistant", 
                "content": response
            }

            if request.type == "smartDebug":
                # Remove inner thoughts from the response
                response = remove_inner_thoughts_from_message(response)
                display_message["content"] = response

                # Modify reply so the display message in the frontend is also have inner thoughts removed
                reply.items[0] = CompletionItem(content=response, isIncomplete=reply.items[0].isIncomplete)

            message_history.append_message(ai_optimized_message, display_message)


        self.reply(reply)
        
        latency_ms = round((time.time() - start) * 1000)
        self.log.info(f"Completion handler resolved in {latency_ms} ms.")

    async def _handle_stream_request(self, request: CompletionRequest, prompt_type: str, model: str) -> None:
        """Handle stream completion request."""
        start = time.time()

        # Use a string buffer to accumulate the full response from streaming chunks.
        # We need to accumulate the response on the backend so that we can save it to
        # the message history after the streaming is complete.
        accumulated_response = ""
        async for reply in self._llm.stream_completions(request, prompt_type, model):
            if isinstance(reply, CompletionStreamChunk):
                accumulated_response += reply.chunk.content

            self.reply(reply)
        
        if request.type != "inline_completion":
            message: ChatCompletionMessageParam = {
                "role": "assistant", 
                "content": accumulated_response
            }
            message_history.append_message(message, message)
        latency_ms = round((time.time() - start) * 1000)
        self.log.info(f"Completion streaming completed in {latency_ms} ms.")

    def reply(self, reply: Any) -> None:
        """Write a reply object to the WebSocket connection.

        Args:
            reply: The completion reply object.
                It must be a dataclass instance.
        """
        message = asdict(reply)
        super().write_message(message)

    def _send_error(self, change: Dict[str, Optional[CompletionError]]) -> None:
        """Send an error message to the client."""
        error = change["new"]

        self.reply(
            ErrorMessage(**asdict(error))
            if error is not None
            else ErrorMessage(error_type="", title="No error", traceback="")
        )



import os
from datetime import datetime
from unittest.mock import patch
from mito_ai.providers import OpenAIProvider
from mito_ai.utils.open_ai_utils import OPEN_SOURCE_AI_COMPLETIONS_LIMIT

REALLY_OLD_DATE = "2020-01-01"
TODAY = datetime.now().strftime("%Y-%m-%d")


def test_os_user_mito_server_below_limit() -> None:
    """
    Open source user, with no OpenAI API key set (Mito server), below both limits.
    No error should be thrown.
    """

    llm = OpenAIProvider()

    with (
        patch.dict(os.environ, {"OPENAI_API_KEY": ""}),
        patch("mito_ai.providers._num_usages", 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
    ):
        capabilities = llm.capabilities

        assert capabilities.provider == "Mito server"
        assert llm.last_error is None


def test_os_user_mito_server_above_limit() -> None:
    """
    Open source user, with no OpenAI API key set (Mito server), above both limits.
    An error should be thrown.
    """

    llm = OpenAIProvider()

    # Above the chat limit
    with (
        patch.dict(os.environ, {"OPENAI_API_KEY": ""}),
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
    ):
        capabilities = llm.capabilities

        assert capabilities.provider == "Mito server"
        assert llm.last_error is not None
        assert llm.last_error.title == "mito_server_free_tier_limit_reached"

    # Above the inline limit
    with (
        patch.dict(os.environ, {"OPENAI_API_KEY": ""}),
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT),
        patch("mito_ai.providers._first_usage_date", REALLY_OLD_DATE),
    ):
        capabilities = llm.capabilities

        assert capabilities.provider == "Mito server"
        assert llm.last_error is not None
        assert llm.last_error.title == "mito_server_free_tier_limit_reached"


def test_os_user_openai_key_set_below_limit() -> None:
    """
    Open source user, with OpenAI API key set, below both limits.
    No error should be thrown.
    """

    llm = OpenAIProvider()

    with (
        patch("mito_ai.providers._num_usages", 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
    ):
        capabilities = llm.capabilities

        assert "user key" in capabilities.provider
        assert llm.last_error is None


def test_os_user_openai_key_set_above_limit() -> None:
    """
    Open source user, with OpenAI API key set, above both limits.
    No error should be thrown, since the user is using their own key.
    """

    llm = OpenAIProvider()

    # Above the chat limit
    with (
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
    ):
        capabilities = llm.capabilities

        assert "user key" in capabilities.provider
        assert llm.last_error is None

    # Above the inline limit
    with (
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT),
        patch("mito_ai.providers._first_usage_date", REALLY_OLD_DATE),
    ):
        capabilities = llm.capabilities

        assert "user key" in capabilities.provider
        assert llm.last_error is None


def test_pro_user_mito_server_set_below_limit() -> None:
    """
    Pro user, with no OpenAI API key set (Mito server), below chat limit.
    No error should be thrown.
    """

    llm = OpenAIProvider()

    with (
        patch.dict(os.environ, {"OPENAI_API_KEY": ""}),
        patch("mito_ai.providers._num_usages", 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
        patch("mito_ai.utils.open_ai_utils.is_pro", return_value=True),
    ):
        capabilities = llm.capabilities

        assert capabilities.provider == "Mito server"
        assert llm.last_error is None


def test_pro_user_mito_server_above_limit() -> None:
    """
    Pro user, with no OpenAI API key set (Mito server), with usage above both limits.
    No error should be thrown since pro users don't have limits.
    """

    llm = OpenAIProvider()

    # Above the chat limit
    with (
        patch.dict(os.environ, {"OPENAI_API_KEY": ""}),
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
        patch("mito_ai.utils.open_ai_utils.is_pro", return_value=True),
    ):
        capabilities = llm.capabilities

        assert capabilities.provider == "Mito server"
        assert llm.last_error is None

    # Above the inline limit
    with (
        patch.dict(os.environ, {"OPENAI_API_KEY": ""}),
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT),
        patch("mito_ai.providers._first_usage_date", REALLY_OLD_DATE),
        patch("mito_ai.utils.open_ai_utils.is_pro", return_value=True),
    ):
        capabilities = llm.capabilities

        assert capabilities.provider == "Mito server"
        assert llm.last_error is None


def test_pro_user_openai_key_set_below_limit() -> None:
    """
    Pro user, with OpenAI API key set, below chat limit.
    No error should be thrown.
    """

    llm = OpenAIProvider()

    with (
        patch("mito_ai.providers._num_usages", 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
        patch("mito_ai.utils.open_ai_utils.is_pro", return_value=True),
    ):
        capabilities = llm.capabilities

        assert "user key" in capabilities.provider
        assert llm.last_error is None


def test_pro_user_openai_key_set_above_limit() -> None:
    """
    Pro user, with OpenAI API key set, above both limits.
    No error should be thrown since pro users don't have limits.
    """

    llm = OpenAIProvider()

    # Above the chat limit
    with (
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1),
        patch("mito_ai.providers._first_usage_date", TODAY),
        patch("mito_ai.utils.open_ai_utils.is_pro", return_value=True),
    ):
        capabilities = llm.capabilities

        assert "user key" in capabilities.provider
        assert llm.last_error is None

    # Above the inline limit
    with (
        patch("mito_ai.providers._num_usages", OPEN_SOURCE_AI_COMPLETIONS_LIMIT),
        patch("mito_ai.providers._first_usage_date", REALLY_OLD_DATE),
        patch("mito_ai.utils.open_ai_utils.is_pro", return_value=True),
    ):
        capabilities = llm.capabilities

        assert "user key" in capabilities.provider
        assert llm.last_error is None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.

import json
from typing import Any, Dict, List, Optional, Type, Final, Union
from datetime import datetime, timedelta

from pydantic import BaseModel
from tornado.httpclient import AsyncHTTPClient

from .db import get_user_field
from .schema import UJ_STATIC_USER_ID, UJ_USER_EMAIL
from .telemetry_utils import (
    MITO_SERVER_FREE_TIER_LIMIT_REACHED,
    log,
)
from .version_utils import is_pro
from openai.types.chat import ChatCompletionMessageParam

MITO_AI_PROD_URL: Final[str] = "https://ogtzairktg.execute-api.us-east-1.amazonaws.com/Prod/completions/"
MITO_AI_DEV_URL: Final[str] = "https://x0l7hinm12.execute-api.us-east-1.amazonaws.com/Prod/completions/"

# If you want to test the dev endpoint, change this to MITO_AI_DEV_URL.
# Note that we have a pytest that ensures that the MITO_AI_URL is always set to MITO_AI_PROD_URL 
# before merging into dev because we always want our users to be using the prod endpoint!
MITO_AI_URL: Final[str] = MITO_AI_PROD_URL

OPEN_SOURCE_AI_COMPLETIONS_LIMIT: Final[int] = 500
OPEN_SOURCE_INLINE_COMPLETIONS_LIMIT: Final[int] = 30 # days

__user_email: Optional[str] = None
__user_id: Optional[str] = None


def check_mito_server_quota(n_counts: int, first_usage_date: str) -> None:
    """Check whether the user has reached the limit of completions for the free tier or not.

    Args:
        n_counts: The number of completions the user has made so far.
        first_usage_date: The date of the user's first usage.
    Raises:
        PermissionError: If the user has reached the limit.
    """
    pro = is_pro()

    if pro:
        return

    if n_counts >= OPEN_SOURCE_AI_COMPLETIONS_LIMIT:
        log(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
        raise PermissionError(MITO_SERVER_FREE_TIER_LIMIT_REACHED)

    if first_usage_date != "":
        first_use = datetime.strptime(first_usage_date, "%Y-%m-%d")
        one_month_later = first_use + timedelta(days=OPEN_SOURCE_INLINE_COMPLETIONS_LIMIT)
        if datetime.now() > one_month_later:
            log(MITO_SERVER_FREE_TIER_LIMIT_REACHED)
            raise PermissionError(MITO_SERVER_FREE_TIER_LIMIT_REACHED)


async def get_ai_completion_from_mito_server(
    last_message_content: Union[str, None],
    ai_completion_data: Dict[str, Any],
    timeout: int,
    max_retries: int,
    n_counts: int,
    first_usage_date: str,
) -> str:
    global __user_email, __user_id

    if __user_email is None:
        __user_email = get_user_field(UJ_USER_EMAIL)
    if __user_id is None:
        __user_id = get_user_field(UJ_STATIC_USER_ID)

    check_mito_server_quota(n_counts, first_usage_date)

    data = {
        "timeout": timeout,
        "max_retries": max_retries,
        "email": __user_email,
        "user_id": __user_id,
        "data": ai_completion_data,
        "user_input": last_message_content or "",  # We add this just for logging purposes
    }

    headers = {
        "Content-Type": "application/json",
    }

    http_client = AsyncHTTPClient(defaults=dict(user_agent="Mito-AI client"))
    try:
        res = await http_client.fetch(
            # Important: DO NOT CHANGE MITO_AI_URL. If you want to use the dev endpoint, 
            # go to the top of this file and change MITO_AI_URL to MITO_AI_DEV_URL. We 
            # have a pytest that ensures that the MITO_AI_URL is always set to MITO_AI_PROD_URL 
            # before merging into dev. So if you change which variable we are using here, the 
            # test will not catch our mistakes.
            MITO_AI_URL, method="POST", headers=headers, body=json.dumps(data)
        )
    finally:
        http_client.close()

    # The lambda function returns a dictionary with a completion entry in it,
    # so we just return that.
    content = json.loads(res.body)
    
    if "completion" in content:
        return content["completion"] # type: ignore
    elif "error" in content:
        raise Exception(f"{content['error']}")
    else:
        raise Exception(f"No completion found in response: {content}")


def get_open_ai_completion_function_params(
    model: str, 
    messages: List[ChatCompletionMessageParam], 
    stream: bool,
    response_format: Optional[Type[BaseModel]] = None,
) -> Dict[str, Any]:
    
    completion_function_params = {
        "model": model,
        "stream": stream,
        "messages": messages,
    }
    
    # If a response format is provided, we need to convert it to a json schema.
    # Pydantic models are supported by the OpenAI API, however, we need to be able to 
    # serialize it for requests that are going to be sent to the mito server. 
    # OpenAI expects a very specific schema as seen below. 
    if response_format:
        json_schema = response_format.schema()
        completion_function_params["response_format"] = {
            "type": "json_schema",
            "json_schema": {
                "name": "plan_of_attack",
                "schema": {
                    **json_schema,
                    "additionalProperties": False
                },
                "strict": True
            }
        }
    
    # o3-mini will error if we try setting the temperature
    if model == "gpt-4o-mini":
        completion_function_params["temperature"] = 0.0

    return completion_function_params


import json
from typing import Any, Dict, Optional
from .version_utils import MITOSHEET_HELPER_PRIVATE, is_pro
from .schema import UJ_MITOSHEET_TELEMETRY, UJ_STATIC_USER_ID, UJ_USER_EMAIL, UJ_FEEDBACKS_V2
from .db import get_user_field
from .._version import __version__
from .utils import is_running_test

import analytics
WRITE_KEY = '6I7ptc5wcIGC4WZ0N1t0NXvvAbjRGUgX' 
analytics.write_key = WRITE_KEY

# If you want, you can optionally choose to print logs
# helpful for debugging.
PRINT_LOGS = False

################################# 
# Mito AI Completion 
# Constants for logging the success or error of Mito AI
MITO_AI_COMPLETION_SUCCESS = 'mito_ai_success'
MITO_AI_COMPLETION_ERROR = 'mito_ai_error'

# Params 
# - logging the type of key 
KEY_TYPE_PARAM = 'AI_key_type'
MITO_SERVER_KEY= 'mito_server_key'
USER_KEY = 'user_key'

# - logging the number of usages of the Mito server
MITO_SERVER_NUM_USAGES = 'mito_server_num_usages'
#################################

################################# 
# Mito Server Free Tier Reached
MITO_SERVER_FREE_TIER_LIMIT_REACHED = 'mito_server_free_tier_limit_reached'
#################################

def telemetry_turned_on() -> bool:
    """
    Helper function that tells you if logging is turned on or
    turned off on the entire Mito instance
    """
    # If private helper is installed, then we don't log anything
    if MITOSHEET_HELPER_PRIVATE:
        return False

    # TODO: Check if the an enterprise user has turned telemetry to true

    # If Mito Pro is on, then don't log anything
    if is_pro():
        return False

    telemetry = get_user_field(UJ_MITOSHEET_TELEMETRY)
    if telemetry is None:
        return False
    
    return bool(telemetry)

def identify() -> None:
    """
    Helper function for identifying a user. We just take
    their python version, mito version, and email.
    """
    if not telemetry_turned_on():
        return

    static_user_id = get_user_field(UJ_STATIC_USER_ID)
    user_email = get_user_field(UJ_USER_EMAIL)
    feedbacks_v2 = get_user_field(UJ_FEEDBACKS_V2)

    params = {
        'version_mitoai': __version__,
        'email': user_email,
        UJ_FEEDBACKS_V2: feedbacks_v2
    }

    if not is_running_test():
        # TODO: If the user is in JupyterLite, we need to do some extra work.
        # You can see this in the mitosheet package. 
        try:
            analytics.identify(static_user_id, params)
        except Exception as e:
            pass

def chunk_param(param: str, param_name: str, chunk_size: int=250) -> Dict[str, str]:
    """
    Split a string into chunks of 250 characters.
    
    Args:
        param: The string to be chunked
        param_name: The name of the param to be chunked (used as prefix for the chunked keys)
        chunk_size: The number of characters in each chunk

    Returns:
        dict: A dictionary with keys 'response_part_1', 'response_part_2', etc.
    """

    chunks = {}

    if not param:
        return {}
    
    num_chunks = (len(param) + chunk_size - 1) // chunk_size

    for i in range(num_chunks):
        start = i * chunk_size
        end = min(start + chunk_size, len(param))
        chunks[f'{param_name}_part_{i + 1}'] = param[start:end]

    return chunks

def log(
        log_event: str, 
        params: Optional[Dict[str, Any]]=None, 
        error: Optional[BaseException]=None, 
    ) -> None:
    """
    This function is the entry point for all logging. 

    If telemetry is not turned off and we are not running tests,
    we log the ai event
    """

    final_params: Dict[str, Any] = params or {}
    
    # Then, make sure to add the user email
    final_params['email'] = get_user_field(UJ_USER_EMAIL)

    # Add the error if it exists
    if error is not None:
        final_params['error'] = str(error)

    # Finally, do the acutal logging. We do not log anything when tests are
    # running, or if telemetry is turned off
    if not is_running_test() and telemetry_turned_on():
        # TODO: If the user is in JupyterLite, we need to do some extra work.
        # You can see this in the mitosheet package. 
        try:
            analytics.track(
                get_user_field(UJ_STATIC_USER_ID), 
                log_event, 
                final_params
            )
        except Exception as e:
            pass
        

    # If we want to print the logs for debugging reasons, then we print them as well
    if PRINT_LOGS:
        print(
            log_event, 
            final_params
        )

    # TODO: Eventually we want to hook this up to the mito log uploader 
    # so enterprises can log usage if they want to.

def log_ai_completion_success(
    key_type: str,
    prompt_type: str,
    last_message_content: str,
    response: Dict[str, Any],
    num_usages: Optional[int] = None
) -> None:
    """
    Logs AI completion success based on the input location.

    Args:
        key_type: The type of key that was used to get the AI completion
        prompt_type: The type of prompt that was sent to the AI
        last_message_content: The last message sent to the AI
        response: The response received from the AI
    """

    # Params that every log has
    base_params = {
        KEY_TYPE_PARAM: key_type,
    }

    try:
        code_cell_input = json.dumps(
            last_message_content.split("Code in the active code cell:")[-1]
            .strip()
            .split("```python")[1]
            .strip()
            .split("```")[0]
        )
    except:
        # Most user prompts will have an associated code cell that serves as the input context.
        # However, types like agent:planning do not have a code cell input.
        code_cell_input = ""

    # Chunk certain params to work around mixpanel's 255 character limit
    code_cell_input_chunks = chunk_param(code_cell_input, "code_cell_input")
    full_prompt_chunks = chunk_param(last_message_content, "full_prompt")
    response_chunks = chunk_param(response["completion"], "response")

    for chunk_key, chunk_value in code_cell_input_chunks.items():
        base_params[chunk_key] = chunk_value

    for chunk_key, chunk_value in full_prompt_chunks.items():
        base_params[chunk_key] = chunk_value

    for chunk_key, chunk_value in response_chunks.items():
        base_params[chunk_key] = chunk_value

    # Log number of usages (for mito server)
    if num_usages is not None:
        base_params[MITO_SERVER_NUM_USAGES] = str(num_usages)

    if prompt_type == "smartDebug":
        error_message = (
            last_message_content.split("Error Message:")[-1]
            .split("ERROR ANALYSIS:")[0]
            .strip()
        )
        error_type = error_message.split(": ")[0]

        final_params = base_params
        final_params["error_message"] = error_message
        final_params["error_type"] = error_type

        log("mito_ai_smart_debug_success", params=final_params)
    elif prompt_type == "codeExplain":
        final_params = base_params

        log("mito_ai_code_explain_success", params=final_params)
    elif prompt_type == "chat":
        final_params = base_params

        # Chunk the user input
        user_input = last_message_content.split("Your task: ")[-1]
        user_input_chunks = chunk_param(user_input, "user_input")
        
        for chunk_key, chunk_value in user_input_chunks.items():
            final_params[chunk_key] = chunk_value

        log("mito_ai_chat_success", params=final_params)
    elif prompt_type == "agent:planning":
        final_params = base_params

        # Chunk the user input
        user_input = last_message_content.split("Your task: ")[-1]
        user_input_chunks = chunk_param(user_input, "user_input")
        
        for chunk_key, chunk_value in user_input_chunks.items():
            final_params[chunk_key] = chunk_value

        log("mito_ai_agent_planning_success", params=final_params)
    elif prompt_type == "agent:execution":
        final_params = base_params

        # Chunk the user input
        user_input = last_message_content.split("Your task: ")[-1]
        user_input_chunks = chunk_param(user_input, "user_input")
        
        for chunk_key, chunk_value in user_input_chunks.items():
            final_params[chunk_key] = chunk_value

        log("mito_ai_agent_execution_success", params=final_params)
    elif prompt_type == "inline_completion":
        final_params = base_params
        log("mito_ai_inline_completion_success", params=final_params)
    else:
        final_params = base_params
        final_params["note"] = (
            "This input_location has not been accounted for in `telemetry_utils.py`."
        )

        log(f"mito_ai_{prompt_type}_success", params=final_params)



import uuid
import os

def get_random_id() -> str:
    """
    Creates a new random ID for the user, which for any given user,
    should only happen once.
    """
    return str(uuid.uuid1())

def is_running_test() -> bool:
    """
    A helper function that quickly returns if the current code is running 
    inside of a test, which is useful for making sure we don't generate 
    tons of logs.
    """
    # Pytest injects PYTEST_CURRENT_TEST into the current environment when running
    running_pytests = "PYTEST_CURRENT_TEST" in os.environ
    # Github injects CI into the environment when running
    running_ci = 'CI' in os.environ and os.environ['CI'] is not None

    return running_pytests or running_ci


import os
import json
from threading import Lock
from typing import Dict, List

from openai.types.chat import ChatCompletionMessageParam
from mito_ai.utils.schema import MITO_FOLDER

CHAT_HISTORY_VERSION = 1 # Increment this if the schema changes

class GlobalMessageHistory:
    """
    Manages a global message history with thread-safe access and persistence.

    This class ensures thread-safe operations for reading, writing, and 
    modifying message histories using a Lock object. It supports loading 
    from and saving to disk, appending new messages, clearing histories, 
    and truncating histories. The histories are stored on disk for persistence.

    Thread safety is crucial to prevent data corruption and race conditions 
    when multiple threads access or modify the message histories concurrently.

    We store two types of messages: AI-optimized and display-optimized messages.
    We store display_history to be able to restore them in the frontend when 
    the extension loads. We store ai_optimized_history to keep the conversation context
    for continuing the conversation.

    The JSON file structure for storing the histories is as follows:
    {
      "chat_history_version": 1,
      "ai_optimized_history": [
        {
          "role": "user",
          "content": "..."
        },
        {
          "role": "assistant",
          "content": "..."
        }
      ],
      "display_history": [
        {
          "role": "user",
          "content": "..."
        },
        {
          "role": "assistant",
          "content": "..."
        }
      ]
    }

    JSON schema is chosen to match what is stored in the memory, so it is 
    easy to serialize/deserialize.

    Attributes:
        _lock (Lock): Ensures thread-safe access.
        _ai_optimized_history (List[Dict[str, str]]): Stores LLM messages.
        _display_history (List[Dict[str, str]]): Stores display messages.
        _save_file (str): Path to the file where histories are saved.

    Methods:
        get_histories() -> tuple[List[Dict[str, str]], List[Dict[str, str]]]:
            Returns copies of the LLM and display histories.
        clear_histories() -> None:
            Clears both histories and saves the changes to disk.
        append_message(ai_optimized_message: Dict[str, str], display_message: Dict[str, str]) -> None:
            Appends new messages to the histories and saves to disk.
        truncate_histories(index: int) -> None:
            Truncates both histories at the given index and saves to disk.
    """
    def __init__(self, save_file: str = os.path.join(MITO_FOLDER, "message_history.json")):
        self._lock = Lock()
        self._ai_optimized_history: List[ChatCompletionMessageParam] = []
        self._display_history: List[ChatCompletionMessageParam] = []
        self._save_file = save_file

        # Load from disk on startup
        self._load_from_disk()
    
    def _load_from_disk(self) -> None:
        """Load existing history from disk, if it exists."""
        if os.path.exists(self._save_file):
            try:
                with open(self._save_file, "r", encoding="utf-8") as f:
                    data = json.load(f)

                    # Check version
                    file_version = data.get("chat_history_version", 0)
                    if file_version == CHAT_HISTORY_VERSION:
                        self._ai_optimized_history = data.get("ai_optimized_history", [])
                        self._display_history = data.get("display_history", [])
                    else:
                        # If versions don't match, delete the file
                        print(
                            f"Warning: Incompatible chat history version ({file_version}). "
                            f"Expected version {CHAT_HISTORY_VERSION}. Deleting file."
                        )
                        f.close()
                        os.remove(self._save_file)
            except Exception as e:
                print(f"Error loading history file: {e}")
    
    def _save_to_disk(self) -> None:
        """Save current history to disk."""
        data = {
            "chat_history_version": CHAT_HISTORY_VERSION,
            "ai_optimized_history": self._ai_optimized_history,
            "display_history": self._display_history,
        }
        # Using a temporary file and rename for safer "atomic" writes
        tmp_file = f"{self._save_file}.tmp"
        try:
            with open(tmp_file, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2)
            os.replace(tmp_file, self._save_file)
        except Exception as e:
            # log or handle error
            print(f"Error saving history file: {e}")

    def get_histories(self) -> tuple[List[ChatCompletionMessageParam], List[ChatCompletionMessageParam]]:
        with self._lock:
            return self._ai_optimized_history[:], self._display_history[:]

    def clear_histories(self) -> None:
        with self._lock:
            self._ai_optimized_history = []
            self._display_history = []
            self._save_to_disk()

    def append_message(self, ai_optimized_message: ChatCompletionMessageParam, display_message: ChatCompletionMessageParam) -> None:
        with self._lock:
            self._ai_optimized_history.append(ai_optimized_message)
            self._display_history.append(display_message)
            self._save_to_disk()

    def truncate_histories(self, index: int) -> None:
        with self._lock:
            self._ai_optimized_history = self._ai_optimized_history[:index]
            self._display_history = self._display_history[:index]
            self._save_to_disk()



from typing import List


def create_chat_prompt(
    variables: List[str],
    active_cell_code: str, 
    input: str
) -> str:
    variables_str = '\n'.join([f"{variable}" for variable in variables])
    prompt = f"""You are an expert python programmer writing a script in a Jupyter notebook. You are given a set of variables, existing code, and a task.

There are two possible types of responses you might give:
1. Code Update: If the task requires modifying or extending the existing code, respond with the updated active code cell and a short explanation of the changes made. 
2. Explanation/Information: If the task does not require a code update, provide an explanation, additional information about a package, method, or general programming question, without writing any code. Keep your response concise and to the point.

When responding:
- Do not use the word "I"
- Do not recreate variables that already exist
- Keep as much of the original code as possible

<Example>

Defined Variables:
{{
    'loan_multiplier': 1.5,
    'sales_df': pd.DataFrame({{
        'transaction_date': ['2024-01-02', '2024-01-02', '2024-01-02', '2024-01-02', '2024-01-03'],
        'price_per_unit': [10, 9.99, 13.99, 21.00, 100],
        'units_sold': [1, 2, 1, 4, 5],
        'total_price': [10, 19.98, 13.99, 84.00, 500]
    }})
}}

Code in the active code cell:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
```

Your task: convert the transaction_date column to datetime and then multiply the total_price column by the sales_multiplier.

Output:
```python
import pandas as pd
sales_df = pd.read_csv('./sales.csv')
sales_df['transaction_date'] = pd.to_datetime(sales_df['transaction_date'])
sales_df['total_price'] = sales_df['total_price'] * sales_multiplier
```

Converted the `transaction_date` column to datetime using the built-in pd.to_datetime function and multiplied the `total_price` column by the `sales_multiplier` variable.

</Example>

Defined Variables:
{variables_str}

Code in the active code cell:
```python
{active_cell_code}
```

Your task: {input}
"""
    
    return prompt


import pytest
from datetime import datetime
from unittest.mock import patch
from mito_ai.utils.open_ai_utils import (
    MITO_AI_PROD_URL,
    MITO_AI_URL,
    MITO_SERVER_FREE_TIER_LIMIT_REACHED,
    OPEN_SOURCE_AI_COMPLETIONS_LIMIT,
    check_mito_server_quota,
)

REALLY_OLD_DATE = "2020-01-01"
TODAY = datetime.now().strftime("%Y-%m-%d")


def test_check_mito_server_quota_open_source_user() -> None:
    # Under both limits
    check_mito_server_quota(1, TODAY)

    # Over the both limits
    with pytest.raises(PermissionError, match=MITO_SERVER_FREE_TIER_LIMIT_REACHED):
        check_mito_server_quota(OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1, REALLY_OLD_DATE)

    # Over the chat limit
    with pytest.raises(PermissionError, match=MITO_SERVER_FREE_TIER_LIMIT_REACHED):
        check_mito_server_quota(OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1, TODAY)

    # Over the inline limit
    with pytest.raises(PermissionError, match=MITO_SERVER_FREE_TIER_LIMIT_REACHED):
        check_mito_server_quota(1, REALLY_OLD_DATE)


def test_check_mito_server_quota_pro_user() -> None:
    # No error should be thrown since pro users don't have limits
    with patch("mito_ai.utils.open_ai_utils.is_pro", return_value=True):
        check_mito_server_quota(1, TODAY)
        check_mito_server_quota(OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1, REALLY_OLD_DATE)
        check_mito_server_quota(OPEN_SOURCE_AI_COMPLETIONS_LIMIT + 1, TODAY)
        check_mito_server_quota(1, REALLY_OLD_DATE)

def test_mito_ai_url_is_prod_url() -> None:
    assert MITO_AI_URL == MITO_AI_PROD_URL



from __future__ import annotations

import traceback
from dataclasses import dataclass, field
from typing import List, Literal, Optional, Type, Union

from pydantic import BaseModel
from openai.types.chat import ChatCompletionMessageParam

from mito_ai.prompt_builders.chat_prompt import create_chat_prompt
from mito_ai.prompt_builders.inline_completer_prompt import create_inline_prompt
from mito_ai.prompt_builders.explain_code_prompt import create_explain_code_prompt
from mito_ai.prompt_builders.smart_debug_prompt import create_error_prompt
from mito_ai.prompt_builders.agent_planning_prompt import create_agent_prompt

CompletionIncomingMessageTypes = Literal['chat', 'inline_completion', 'codeExplain', 'smartDebug', 'agent:planning']
IncomingMessageTypes = Union[Literal['clear_history', 'fetch_history'], CompletionIncomingMessageTypes]

@dataclass(frozen=True)
class AICapabilities:
    """
    AI provider capabilities
    """

    # Configuration schema.
    configuration: dict

    # AI provider name.
    provider: str

    # Message type.
    type: str = "ai_capabilities"

@dataclass(frozen=True)
class ChatMessageBuilder:
    variables: Optional[List[str]] = None
    activeCellCode: Optional[str] = None
    input: Optional[str] = None
    index: Optional[int] = None

    @property
    def prompt(self) -> str:
        return create_chat_prompt(self.variables or [], self.activeCellCode or '', self.input or '')
    
    @property
    def display_message(self) -> str:
        cell_code_block = f"""```python
{self.activeCellCode}
```

"""
        return f"{cell_code_block if self.activeCellCode else ''}{self.input}"
    
    @property
    def pro_model(self) -> str:
        return "o3-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class SmartDebugMessageBuilder:
    variables: Optional[List[str]] = None
    activeCellCode: Optional[str] = None
    errorMessage: Optional[str] = None

    @property
    def prompt(self) -> str:
        return create_error_prompt(self.errorMessage or '', self.activeCellCode or '', self.variables or [])
    
    @property
    def display_message(self) -> str:
        cell_code_block = f"""```python
{self.activeCellCode}
```

"""
        return f"{cell_code_block if self.activeCellCode else ''}{self.errorMessage}"
    
    @property
    def pro_model(self) -> str:
        return "gpt-4o-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class CodeExplainMessageBuilder:
    variables: Optional[List[str]] = None
    activeCellCode: Optional[str] = None

    @property
    def prompt(self) -> str:
        return create_explain_code_prompt(self.activeCellCode or '')
    
    @property
    def display_message(self) -> str:
        cell_code_block = f"""```python
{self.activeCellCode}
```

"""
        
        return f"{cell_code_block if self.activeCellCode else ''}Explain this code"
    
    @property
    def pro_model(self) -> str:
        return "gpt-4o-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class InlineCompletionMessageBuilder:
    prefix: Optional[str] = None
    suffix: Optional[str] = None
    variables: Optional[List[str]] = None

    @property
    def prompt(self) -> str:
        return create_inline_prompt(self.prefix or '', self.suffix or '', self.variables or [])
    
    @property
    def pro_model(self) -> str:
        return "gpt-4o-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

@dataclass(frozen=True)
class AgentMessageBuilder:
    fileType: Optional[str] = None
    columnSamples: Optional[List[str]] = None
    input: Optional[str] = None
    variables: Optional[List[str]] = None

    @property
    def prompt(self) -> str:
        return create_agent_prompt(
            self.fileType or "",
            self.columnSamples or [],
            self.input or "",
            self.variables or [],
        )

    @property
    def display_message(self) -> str:
        return self.input or ''
    
    @property
    def pro_model(self) -> str:
        return "o3-mini"
    
    @property
    def os_model(self) -> str:
        return "gpt-4o-mini"

    @property
    def response_format(self) -> Type[BaseModel]:
        class PlanOfAttack(BaseModel):
            actions: List[str]
            dependencies: List[str]
        return PlanOfAttack
    


@dataclass(frozen=True)
class CompletionRequest:
    """
    Message send by the client to request an AI chat response.
    """

    # Message type.
    type: IncomingMessageTypes

    # Message UID generated by the client.
    message_id: str

    # Chat messages.
    messages: List[ChatCompletionMessageParam] = field(default_factory=list)

    # Whether to stream the response (if supported by the model).
    stream: bool = False


@dataclass(frozen=True)
class CompletionItemError:
    """
    Completion item error information.
    """

    # Error message.
    message: Optional[str] = None


@dataclass(frozen=True)
class CompletionItem:
    """
    A completion suggestion.
    """

    # The completion.
    content: str

    # Whether the completion is incomplete or not.
    isIncomplete: Optional[bool] = None
    
    # Unique token identifying the completion request in the frontend.
    token: Optional[str] = None

    # Error information for the completion item.
    error: Optional[CompletionItemError] = None


@dataclass(frozen=True)
class CompletionError:
    """
    Completion error description.
    """

    # Error type.
    error_type: str

    # Error title.
    title: str

    # Error traceback.
    traceback: str

    # Hint to resolve the error.
    hint: str = ""

    @staticmethod
    def from_exception(exception: BaseException, hint: str = "") -> CompletionError:
        """
        Create a completion error from an exception.
        
        Note: OpenAI exceptions can include a 'body' attribute with detailed error information.
        While mypy doesn't know about this attribute on BaseException, we need to handle it
        to properly extract error messages from OpenAI API responses.
        """
        error_type = type(exception)
        error_module = getattr(error_type, "__module__", "")
        return CompletionError(
            error_type=f"{error_module}.{error_type.__name__}"
            if error_module
            else error_type.__name__,
            title=exception.body.get("message")
            if hasattr(exception, "body")
            else (exception.args[0] if exception.args else "Exception"),
            traceback=traceback.format_exc(),
            hint=hint,
        )


@dataclass(frozen=True)
class ErrorMessage(CompletionError):
    """
    Error message.
    """

    # Message type.
    type: Literal["error"] = "error"



@dataclass(frozen=True)
class CompletionReply:
    """
    Message sent from model to client with the completion suggestions.
    """

    # List of completion items.
    items: List[CompletionItem]

    # Parent message UID.
    parent_id: str

    # Message type.
    type: Literal["reply"] = "reply"

    # Completion error.
    error: Optional[CompletionError] = None


@dataclass(frozen=True)
class CompletionStreamChunk:
    """
    Message sent from model to client with the infill suggestions
    """

    chunk: CompletionItem

    # Parent message UID.
    parent_id: str

    # Whether the completion is done or not.
    done: bool

    # Message type.
    type: Literal["chunk"] = "chunk"

    # Completion error.
    error: Optional[CompletionError] = None
    """Completion error."""

@dataclass(frozen=True)
class FetchHistoryReply:
    """
    Message sent from model to client with the chat history.
    """

    # Message UID.
    parent_id: str

    # List of chat messages.
    items: List[ChatCompletionMessageParam]

    # Message type.
    type: Literal["reply"] = "reply"




from typing import List


def create_agent_prompt(
    file_type: str, columnSamples: List[str], input: str, variables: List[str]
) -> str:
    variables_str = "\n".join([f"{variable}" for variable in variables])

    if file_type:
        file_sample_snippet = f"""You will be working with the following dataset (sample rows shown) from a {file_type} file:
{columnSamples}
"""

    return f"""You are an expert data science assistant working in a Jupyter notebook environment. Your task is to break a problem into the essential, actionable steps required to write Python code for solving it. 
You have access to the following Python packages:
- pandas (for data manipulation and analysis)
- matplotlib (for data visualization)

Defined Variables:
{variables_str}

{file_sample_snippet if file_type else ''}

Given the dataset (if provided) and the question below:
1. Break the problem into the **smallest possible number of clear, high-level tasks** necessary to achieve the solution. 
2. **Do not include any code or specific implementation details.** Focus only on describing the high-level steps required to solve the problem.
3. Additionally, provide a list of python packages that are required to complete the actions. 

Your task: {input}
"""
